import os

DEBUG = True
TEMPLATE_DEBUG = DEBUG
ADMINS = ()
MANAGERS = ADMINS
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'djembe.db',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}
TIME_ZONE = 'America/Chicago'
LANGUAGE_CODE = 'en-us'
SITE_ID = 1
USE_I18N = True
USE_L10N = True
USE_TZ = True
MEDIA_ROOT = ''
MEDIA_URL = ''
STATIC_ROOT = ''
STATIC_URL = '/static/'
SECRET_KEY = 'sa)2m1h$n+g-5#8&amp;x!6dk_0yio_sjc4o7^+t(jp5)h_(m69esw'
INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'djembe'
)
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
        'djembe': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
    }
}
EMAIL_BACKEND = 'djembe.backends.EncryptingTestBackend'
TEST_RUNNER = 'djembe.testrunners.TestSuiteRunner'
COVERAGE_REPORT_HTML_OUTPUT_DIR = os.path.join(os.path.dirname(__file__), '../coverage')
COVERAGE_MODULE_EXCLUDES = [
    'tests$',
    'settings$',
    '__init__',
    'migrations',
    'djembe.admin',
    'djembe.testrunners'
]



from django.core import mail
from django.conf import settings

from django_coverage.coverage_runner import CoverageRunner


class TestSuiteRunner(CoverageRunner):
    """
    Just resets EMAIL_BACKEND to whatever was specified in settings.
    """

    def setup_test_environment(self, **kwargs):
        super(TestSuiteRunner, self).setup_test_environment(**kwargs)
        settings.EMAIL_BACKEND = mail.original_email_backend



from django.contrib import admin

from djembe.models import Identity


class IdentityAdmin(admin.ModelAdmin):
    model = Identity
    list_display = ['address', 'fingerprint']
    search_fields = ['address']
admin.site.register(Identity, IdentityAdmin)



import re

from django.db import models
from django.utils.translation import gettext_lazy as _

from M2Crypto import X509


class Identity(models.Model):
    certificate = models.TextField(
        help_text=_('A PEM-encoded X.509 certificate.')
    )

    address = models.EmailField(
        blank=True,
        max_length=256,
        help_text=_('If left blank, it will be extracted from the X.509 certificate.')
    )

    key = models.TextField(
        blank=True,
        help_text=_('If mail <em>from</em> this identity should be signed, put a PEM-encoded private key here. Make sure it does not require a passphrase.')
    )

    class Meta:
        ordering = ['address']
        verbose_name_plural = _('Identities')

    def __unicode__(self):
        return self.address or self

    @property
    def fingerprint(self):
        fingerprint = self.x509.get_fingerprint(md='sha1').rjust(40, '0')
        return re.sub(r'(..)(?!$)', r'\1:', fingerprint)

    @property
    def x509(self):
        return X509.load_cert_string(str(self.certificate))


def set_identity_address_from_certificate(sender, **kwargs):
    identity = kwargs['instance']
    if not identity.address:
        x509 = identity.x509
        subject = x509.get_subject()
        email_address = subject.get_entries_by_nid(subject.nid['emailAddress'])[0]
        identity.address = str(email_address.get_data())

models.signals.pre_save.connect(set_identity_address_from_certificate, sender=Identity)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Identity'
        db.create_table('djembe_identity', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('address', self.gf('django.db.models.fields.EmailField')(max_length=256, blank=True)),
            ('certificate', self.gf('django.db.models.fields.TextField')()),
            ('key', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal('djembe', ['Identity'])


    def backwards(self, orm):
        # Deleting model 'Identity'
        db.delete_table('djembe_identity')


    models = {
        'djembe.identity': {
            'Meta': {'ordering': "['address']", 'object_name': 'Identity'},
            'address': ('django.db.models.fields.EmailField', [], {'max_length': '256', 'blank': 'True'}),
            'certificate': ('django.db.models.fields.TextField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.TextField', [], {'blank': 'True'})
        }
    }

    complete_apps = ['djembe']


import email
import logging
import smtplib
import sys

from django.conf import settings
from django.core.mail.backends import base
from django.core.mail.backends import smtp
from django.core.mail.message import make_msgid
from django.core.mail.message import sanitize_address

from M2Crypto import BIO
from M2Crypto import SMIME
from M2Crypto import X509

from djembe.models import Identity


class EncryptingBackendMixin(object):
    """
    A mixin to encrypt, possibly sign, and finally deliver messages.

    Backends built with this should implement deliver(), and call send() for
    each message in send_messages().
    """
    logger = logging.getLogger('djembe.backends.EncryptingBackendMixin')

    payload_mime_headers = [
        'Content-Disposition',
        'Content-Type',
        'Mime-Version',
    ]

    def analyze_recipients(self, email_message):
        """
        Determine which recipients should get encrypted messages.
        """
        encrypting_identities = plaintext_recipients = None

        if email_message.recipients():
            recipients = set([
                sanitize_address(addr, email_message.encoding)
                for addr in email_message.recipients()
            ])

            encrypting_identities = Identity.objects.filter(address__in=recipients)
            encrypting_recipients = set([r.address for r in encrypting_identities])
            plaintext_recipients = recipients - encrypting_recipients

        return (encrypting_identities, encrypting_recipients, plaintext_recipients)

    def deliver(self, sender_address, recipients, message):
        """
        Handles the actual delivery of a message.
        """
        raise NotImplementedError

    def encrypt(self, sender_address, encrypting_identities, message):
        """
        Encrypts the given message for all the supplied recipients.
        """

        if not encrypting_identities:
            raise ValueError('Encrypting recipient identities not supplied.')

        if not message:
            raise ValueError('Valid Message not supplied.')

        s = SMIME.SMIME()

        cipher = getattr(settings, 'DJEMBE_CIPHER', 'aes_256_cbc')
        s.set_cipher(SMIME.Cipher(cipher))

        self.logger.debug("Encrypting message for %s" % encrypting_identities)

        # Gather all the recipient certificates
        sk = X509.X509_Stack()
        for identity in encrypting_identities:
            sk.push(identity.x509)
        s.set_x509_stack(sk)

        # prepare the payload for encryption
        payload_msg = self.extract_payload(message)

        # encrypt the payload
        payload = BIO.MemoryBuffer(payload_msg.as_string())
        pkcs7_encrypted_data = s.encrypt(payload)
        payload.close()

        # get the PKCS7 object into a string
        payload = BIO.MemoryBuffer()
        s.write(payload, pkcs7_encrypted_data)
        pkcs7_string = payload.read()
        payload.close()

        encrypted_message = email.message_from_string(pkcs7_string)

        message.set_payload(encrypted_message.get_payload())

        for header, value in encrypted_message.items():
            del message[header]
            message[header] = value
        del message['Message-ID']
        message['Message-ID'] = make_msgid()

        return message

    def extract_payload(self, message):
        payload_msg = email.message.Message()
        payload_msg.set_payload(message.get_payload())
        for header in self.payload_mime_headers:
            payload_msg[header] = message[header]
        return payload_msg

    def get_sender_identity(self, address):
        """
        Looks for an Identity matching the sender address.
        """
        sender_identity = None
        if address:
            sender_identities = Identity.objects.filter(
                address=address,
            ).exclude(key='')
            sender_count = sender_identities.count()
            if sender_count > 0:
                if sender_count == 1:
                    sender_identity = sender_identities[0]
                else:
                    self.logger.warning('Sender matches multiple identities; cannot sign the message.')
        else:
            raise ValueError('Sender address not supplied.')

        return sender_identity

    def send(self, email_message):
        """
        Sends a message, possibly signed, possibly encrypted.

        If the sender has a key, the message will be signed with it.

        Recipients for whom an Identity can be found will be sent an encrypted
        version, any others get plaintext.
        """
        sender_address = sanitize_address(
            email_message.from_email,
            email_message.encoding
        )
        sender_identity = self.get_sender_identity(sender_address)

        encrypting_identities, encrypting_recipients, plaintext_recipients = self.analyze_recipients(email_message)

        # work with the regular standard library message instead of Django's wrapper
        message = email_message.message()

        if sender_identity:
            message = self.sign(sender_identity, message)

        sent = 0
        if plaintext_recipients:
            try:
                self.deliver(
                    sender_address,
                    plaintext_recipients,
                    message.as_string()
                )
                sent += 1
            except:
                if self.fail_silently is False:
                    raise

        if encrypting_identities:
            try:
                encrypted_message = self.encrypt(
                    sender_address,
                    encrypting_identities,
                    message
                )

                self.deliver(
                    sender_address,
                    encrypting_recipients,
                    encrypted_message.as_string()
                )
                sent += 1
            except:
                if self.fail_silently is False:
                    if not sent:
                        raise
                    else:
                        exc_class, exc, tb = sys.exc_info()
                        new_exc = exc_class("Only partial success (messages sent before error: %s)" % sent)
                        raise new_exc.__class__, new_exc, tb

        return sent

    def sign(self, sender_identity, message):
        """
        Signs an email message.
        """

        self.logger.debug('Signing message as %s' % sender_identity)

        s = SMIME.SMIME()

        signing_cert = BIO.MemoryBuffer(str(sender_identity.certificate))
        signing_key = BIO.MemoryBuffer(str(sender_identity.key))
        s.load_key_bio(signing_key, signing_cert)

        # extract the payload from the original message, construct a temporary
        # message without all the header info, and sign that message's string
        # representation
        payload_msg = self.extract_payload(message)
        content_to_sign = payload_msg.as_string()

        pkcs7_signed_data = s.sign(
            BIO.MemoryBuffer(content_to_sign),
            flags=SMIME.PKCS7_DETACHED
        )

        # get the PKCS7 object into a string
        payload = BIO.MemoryBuffer()
        s.write(
            payload,
            pkcs7_signed_data,
            BIO.MemoryBuffer(content_to_sign),
            flags=SMIME.PKCS7_DETACHED
        )
        pkcs7_string = payload.read()
        payload.close()

        signed_message = email.message_from_string(pkcs7_string)

        message.set_payload(signed_message.get_payload())

        for header, value in signed_message.items():
            del message[header]
            message[header] = value

        return message


class EncryptingSMTPBackend(EncryptingBackendMixin, smtp.EmailBackend):
    """
    Delivers encrypted messages via SMTP.
    """

    def deliver(self, sender_address, recipients, message):
        """
        Handles the actual delivery of a message.
        """
        self.logger.info("Delivering message from %s to %s" % (sender_address, recipients))
        return self.connection.sendmail(
            sender_address,
            recipients,
            message
        )

    def send_messages(self, email_messages):
        """
        Sends one or more EmailMessage objects and returns the number of email
        messages sent.
        """
        if not email_messages:
            return 0
        self._lock.acquire()
        try:
            new_conn_created = self.open()
            if not self.connection and self.fail_silently is False:
                raise smtplib.SMTPConnectError('Cannot send without a valid connection')
            num_sent = 0
            for message in email_messages:
                num_sent += self.send(message)
            if new_conn_created:
                self.close()
        finally:
            self._lock.release()
        return num_sent


class EncryptingTestBackend(EncryptingBackendMixin, base.BaseEmailBackend):
    """
    Collects encrypted messages for review, instead of actually delivering them.
    """

    messages = []

    def deliver(self, sender_address, recipients, message):

        # poison pill #1
        if sender_address == 'breakerbreaker@example.com' and 'recipient1@example.com' in recipients:
            raise ValueError("He has a secret CB RADIO!")

        # poison pill #2
        if sender_address == 'breakerofthings@example.com':
            raise ValueError("He just can't help it.")

        self.messages.append({
            'sender': sender_address,
            'recipients': recipients,
            'message': message
        })

    def send_messages(self, email_messages):
        if not email_messages:
            return 0

        num_sent = 0
        for message in email_messages:
            num_sent += self.send(message)

        return num_sent



from django.core import mail
from django.test import TestCase

from djembe.models import Identity
from djembe.tests import data

from M2Crypto import BIO
from M2Crypto import SMIME
from M2Crypto import X509


class EncryptionTest(TestCase):

    def setUp(self):
        self.recipient1 = Identity.objects.create(
            certificate=data.RECIPIENT1_CERTIFICATE,
            key=data.RECIPIENT1_KEY
        )

        self.recipient2 = Identity.objects.create(
            certificate=data.RECIPIENT2_CERTIFICATE
        )

        self.list_member = Identity.objects.create(
            certificate=data.RECIPIENT1_CERTIFICATE,
            address='list@example.com'
        )

        self.list_member = Identity.objects.create(
            certificate=data.RECIPIENT2_CERTIFICATE,
            address='list@example.com'
        )

        self.text_template = 'S/MIME multipart test %s'
        self.html_template = '<h1>S/MIME Test</h1><p>Message <strong>%s</strong></p>'

    def testAllTheThings(self):
        """
        Test the full scenario: multiple encrypted and plaintext recipients.

        Tests that multiple recipients can all read a message, and that
        recipients with no Identity records get plain text.
        """
        count = 1
        sender = Identity.objects.get(address='recipient1@example.com')
        recipients = [identity.address for identity in Identity.objects.all()]
        recipients.extend([
            'recipient3@example.com',
            'recipient4@example.com'
        ])
        message = mail.EmailMultiAlternatives(
            self.text_template % count,
            self.text_template % count,
            sender.address,
            recipients
        )
        message.attach_alternative(self.html_template % count, "text/html")
        message.send()

        backend = mail.get_connection()
        self.assertEqual(len(backend.messages), 2)

        #
        # verify the encryption and signature
        #
        s = SMIME.SMIME()

        # Load the sender's cert.
        x509 = X509.load_cert_string(data.RECIPIENT1_CERTIFICATE)
        sk = X509.X509_Stack()
        sk.push(x509)
        s.set_x509_stack(sk)

        # Load the sender's CA cert.
        st = X509.X509_Store()
        st.add_x509(x509)
        s.set_x509_store(st)

        # Decrypt the message as both encrypted recipients

        #
        # recipient 1
        #
        recipient1_cert = BIO.MemoryBuffer(data.RECIPIENT1_CERTIFICATE)
        recipient1_key = BIO.MemoryBuffer(data.RECIPIENT1_KEY)
        s.load_key_bio(recipient1_key, recipient1_cert)

        msg = BIO.MemoryBuffer(backend.messages[1]['message'])
        p7, msg_data = SMIME.smime_load_pkcs7_bio(msg)
        out = s.decrypt(p7)

        # Verify the message
        msg = BIO.MemoryBuffer(out)
        p7, msg_data = SMIME.smime_load_pkcs7_bio(msg)
        verified_msg = s.verify(p7, msg_data)
        self.assertTrue(verified_msg)

        #
        # recipient 2
        #
        recipient2_cert = BIO.MemoryBuffer(data.RECIPIENT2_CERTIFICATE)
        recipient2_key = BIO.MemoryBuffer(data.RECIPIENT2_KEY)
        s.load_key_bio(recipient2_key, recipient2_cert)

        msg = BIO.MemoryBuffer(backend.messages[1]['message'])
        p7, msg_data = SMIME.smime_load_pkcs7_bio(msg)
        out = s.decrypt(p7)

        # Verify the message
        msg = BIO.MemoryBuffer(out)
        p7, msg_data = SMIME.smime_load_pkcs7_bio(msg)
        self.assertTrue(s.verify(p7, msg_data))

        # verify that the plaintext also got through
        msg = BIO.MemoryBuffer(backend.messages[1]['message'])

    def testEncryptedDeliveryProblem(self):
        subject = 'No! Not the radio!'
        body = "10-4 good buddy!"
        sender = "breakerbreaker@example.com"
        recipient = "recipient1@example.com"

        try:
            mail.send_mail(subject, body, sender, [recipient])
            self.fail("Unless you're four-wheeling, CB radio is a problem.")
        except ValueError:
            pass

    def testIdentityInstance(self):
        self.assertEqual('C6:AF:98:41:75:D4:10:E9:BE:0A:5C:D8:7F:0E:6F:BB:A7:E1:B0:0E', self.recipient1.fingerprint)

    def testMixedMessages(self):
        message1 = mail.message.EmailMessage(
            subject='This is a poison message.',
            body="And will cause an exception.",
            from_email="breakerbreaker@example.com",
            to=["somebody@example.com", "recipient1@example.com"]
        )

        backend = mail.get_connection()

        try:
            backend.send_messages([message1])
            self.fail("Poison message should have thrown an exception.")
        except ValueError:
            pass

    def testNoMessageToEncrypt(self):
        backend = mail.get_connection()
        try:
            backend.encrypt('recipient1@example.com', ['recipient2@example.com'], '')
            self.fail('Lack of recipients should have raised an exception from encrypt method.')
        except ValueError:
            pass

    def testNoMessages(self):
        backend = mail.get_connection()
        sent = backend.send_messages([])
        self.assertEqual(0, sent)

    def testNoRecipientsToEncrypt(self):
        backend = mail.get_connection()
        try:
            backend.encrypt('recipient1@example.com', [], '')
            self.fail('Lack of recipients should have raised an exception from encrypt method.')
        except ValueError:
            pass

    def testPlainTextDeliveryProblem(self):
        subject = 'This is a poison message.'
        body = "And will cause an exception."
        sender = "breakerofthings@example.com"
        recipient = "deadletteroffice@example.com"

        try:
            mail.send_mail(subject, body, sender, [recipient])
            self.fail("Poison message should have thrown an exception.")
        except ValueError:
            pass

    def testSenderIdentity(self):
        backend = mail.get_connection()

        # should get an error with no address
        try:
            sender = backend.get_sender_identity('')
            self.fail('Lack of sender address should have raised an exception')
        except ValueError:
            pass

        # a single valid sender should return a valid Identity
        sender_address = 'multisender@example.com'
        Identity.objects.create(
            certificate=data.RECIPIENT1_CERTIFICATE,
            key=data.RECIPIENT1_KEY,
            address=sender_address
        )
        sender = backend.get_sender_identity(sender_address)
        self.assertTrue(sender is not None)
        self.assertEqual(sender.address, sender_address)

        # No Identity should be returned when more than one is found with a
        # given address
        Identity.objects.create(
            certificate=data.RECIPIENT1_CERTIFICATE,
            key=data.RECIPIENT1_KEY,
            address=sender_address
        )
        sender = backend.get_sender_identity(sender_address)
        self.assertTrue(sender is None)



from setuptools import setup

djembe = __import__('djembe')

with open('README.rst') as file:
    long_description = file.read()

setup(
    author='John Hensley',
    author_email='john@cabincode.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Communications :: Email',
        'Topic :: Security :: Cryptography',
    ],
    description=djembe.__doc__,
    long_description=long_description,
    download_url='http://github.com/cabincode/django-djembe/downloads',
    install_requires=[
        'Django>=1.4',
        'M2Crypto',
        'South'
    ],
    name='django-djembe',
    packages=[
        'djembe',
        'djembe.migrations',
        'djembe.tests'
    ],
    package_data={
        'djembe': [
            'README.rst',
            'LICENSE.txt',
        ],
    },
    tests_require=[
        'Django>=1.4',
        'M2Crypto',
        'South',
        'coverage',
        'django_coverage'
    ],
    test_suite='tests.main',
    url='http://github.com/cabincode/django-djembe',
    version=djembe.get_version(),
)
